import numpy as np
from cvxopt import matrix
from cvxopt.solvers import qp as cvxqp
from cvxopt.solvers import options
from quadprog import solve_qp

from surropt.utils.matrixdivide import mldivide

options['show_progress'] = False  # disable cvxopt output
options['abstol'] = 1e-8
options['feastol'] = 1e-8


# Linesearch auxiliary routine
def __linesearch(x, p, objfun, nonlcon, lambdav, obj, objgrad, globalls):
    eta = 0.25
    tau = 0.5

    # Choose mu to satisfy 18.36 with sigma = 1
    delta_bar = np.sqrt(np.spacing(1))

    if lambdav.size == 0:  # no consraint multipliers
        mu = 1 / delta_bar
    else:
        mu = 1 / (np.linalg.norm(lambdav, np.inf) + delta_bar)

    alpha = 1.0

    c = objgrad

    d, ce, *_ = nonlcon(x)

    # call the merit function
    phi_x_mu, obj, objgrad, globalls = __merit_fun(obj, objgrad, objfun, nonlcon, x, mu, globalls)

    D_phi_x_mu = c.conj().T @ p  # directional derivative

    # only the elements of d corresponding to violated constraints should be included
    idx = d > 0
    if idx.size != 0:
        t = - np.linalg.norm(np.vstack((ce, d[idx].reshape(-1, 1))), 1) / mu
    else:
        t = - np.linalg.norm(ce, 1) / mu

    if t.size != 0:
        D_phi_x_mu += t

    while True:  # while loop to find alpha that causes decrease in merit
        p1, obj, objgrad, globalls = __merit_fun(np.array([]), objgrad, objfun, nonlcon, x + alpha * p, mu, globalls)
        p2 = phi_x_mu + eta * alpha * D_phi_x_mu

        if p1 > p2:
            # Reset alpha = tau_alpha * alpha for some tau_alpha in the range [0, 1]
            tau_alpha = 0.9 * tau
            alpha *= tau_alpha

        else:
            break  # the alpha is enough to cause decrease

    x_new = x + alpha * p

    return x_new, alpha, obj, objgrad, globalls


# Merit function evaluation auxiliary routine
def __merit_fun(obj, objgrad, objfun, nonlcon, x, mu, globalls):
    ci, ce, *_ = nonlcon(x)

    idx = ci > 0  # indexes of negative (violated) constraints

    if idx.size != 0:
        con = np.vstack((ce, ci[idx].reshape(-1, 1)))
    else:
        con = ce

    if type(obj) is np.ndarray and obj.size == 0:
        obj, objgrad = objfun(x)  # evaluate function if obj value is empty
        globalls['nevals'] += 1

    merit = obj
    t = np.linalg.norm(con, 1) / mu

    if t.size != 0:  # if t is not empty
        merit += t

    return merit, obj, objgrad, globalls


# Transform bounds into constraints (gradients included)
def __bnd2cf(x, lbidx, ubidx, lb, ub, bnds_grad, nonlcon):
    cif_eval, cef, cigf_eval, cegf = nonlcon(x)

    if cef.size == 0:
        cef = cef.reshape(-1, 1)  # in case of empty array, reshape it to avert vstack problems

    if cegf.size == 0:
        cegf = cegf.reshape(-1, x.size)  # same reshape logic as cef

    if cif_eval.size != 0:
        cif = np.vstack((cif_eval, lb[lbidx].reshape(-1, 1) - x[lbidx].reshape(-1, 1),
                         x[ubidx].reshape(-1, 1) - ub[ubidx].reshape(-1, 1)))
    else:
        cif = np.vstack((lb[lbidx].reshape(-1, 1) - x[lbidx].reshape(-1, 1),
                         x[ubidx].reshape(-1, 1) - ub[ubidx].reshape(-1, 1)))

        if cif.size == 0:  # boundless problem and without constraints
            cif = cif.reshape(-1, 1)  # to avoid concatenation problem

    if cigf_eval.size != 0:
        cigf = np.vstack((cigf_eval, bnds_grad))
    else:
        cigf = bnds_grad

    return cif, cef, cigf, cegf


# Check inputs auxiliaries (sanitize input vectors by enforcing them to be 1D vectors of same size)
def __check_vector_input(x0, lb, ub):
    # if any of the bounds is set to None, make them a infinity vector
    if lb is None:
        lb = - np.ones((x0.size, )) * np.inf
    else:
        lb = lb.flatten()

    if ub is None:
        ub = np.ones((x0.size, )) * np.inf
    else:
        ub = ub.flatten()

    x0 = x0.flatten()  # force to be 1D array

    if x0.size != lb.size or x0.size != ub.size:
        ValueError("x0, lb and ub has to have the same number of elements.")

    if np.any(lb >= ub):
        ValueError("All lower bounds must be lower than the upper bounds (lb < ub).")

    return x0, lb, ub


def __set_options_structure(options, x0):
    if options is None:
        # set the default values of the options structure
        maxiter = 400
        maxfunevals = 100 * x0.size
        tolstep = np.sqrt(np.spacing(1))
        tolopt = np.sqrt(np.spacing(1))
        tolcon = np.sqrt(np.spacing(1))
        qpsolver = 'quadprog'

        return maxiter, maxfunevals, tolstep, tolopt, tolcon, qpsolver

    elif type(options) is dict:  # user wants to set non-default values
        # check for the keys inside the dictionary
        valid_opts = ['maxiter', 'maxfunevals', 'tolstep', 'tolopt', 'tolcon']

        # get all keys set in dictionary that are invalid
        invalid_opts = [opt for opt in valid_opts if opt not in options]
        if len(invalid_opts) != 0:  # at least one invalid option was set. Throw exception
            ValueError("Invalid options were set. They are: {}".format(invalid_opts))

        # maximum number of iterations
        if 'maxiter' in options:
            maxiter = options['maxiter']
            if not np.isscalar(maxiter) or maxiter <= 0 or np.fix(maxiter) != maxiter:
                ValueError('Maximum number of iterations must be a positive integer scalar value.')
        else:
            maxiter = 400  # set to default

        # maximum number of function evaluations
        if 'maxfunevals' in options:
            maxfunevals = options['maxfunevals']
            if not np.isscalar(maxfunevals) or maxfunevals <= 0 or np.fix(maxfunevals) != maxfunevals:
                ValueError('Maximum number of function evaluations must be a positive integer scalar value.')

        else:
            maxfunevals = 100 * x0.size

        # minimum step size
        if 'tolstep' in options:
            tolstep = options['tolstep']
            if not np.isscalar(tolstep) or tolstep <= 0:
                ValueError('Invalid value set for step size tolerance.')

        else:
            tolstep = 1e-6

        # optimality tolerance
        if 'tolopt' in options:
            tolopt = options['tolopt']
            if not np.isscalar(tolopt) or tolopt <= 0:
                ValueError('Invalid value set for optimality tolerance.')

        else:
            tolopt = 1e-6

        # constraint tolerance
        if 'tolcon' in options:
            tolcon = options['tolcon']
            if not np.isscalar(tolcon) or tolcon <= 0:
                ValueError('Invalid value set for constraint tolerance.')

        else:
            tolcon = 1e-6

        # QP solver
        if 'qpsolver' in options:
            if isinstance(options['qpsolver'], str):  # if it is a string, check for valid options
                qpsolver = options['qpsolver']
                if qpsolver != 'cvxopt' or qpsolver != 'quadprog':
                    ValueError("Invalid QP solver option.")

        else:
            qpsolver = 'quadprog'

        return maxiter, maxfunevals, tolstep, tolopt, tolcon, qpsolver

    else:  # invalid argument
        ValueError("options argument has to be a dictionary.")


def __objective_function_check(objfun, x0: np.ndarray):
    if callable(objfun):
        obj, gradobj = objfun(x0)

        if not isinstance(obj, np.ndarray):
            obj = np.asarray(obj)

        if not isinstance(gradobj, np.ndarray):
            gradobj = np.asarray(gradobj)

        if obj.size != 1:
            ValueError("The objective function evaluation must return a single value (scalar)")

        if gradobj.ndim > 1 or gradobj.size == 0 or gradobj.size != x0.size:
            ValueError("The gradient evaluation has to return a (non-empty) 1D array of the same size as x0.")
    else:
        TypeError("Objective function and its gradient must be a callable function")


def __constraint_function_check(confun, x0):
    if callable(confun):  # if confun is a callable function
        cif_eval, cef_eval, cigf_eval, cegf_eval = confun(x0)

        if not isinstance(cif_eval, np.ndarray):
            cif_eval = np.asarray(cif_eval)  # convert to numpy array

        if not isinstance(cef_eval, np.ndarray):
            cef_eval = np.asarray(cef_eval)

        if not isinstance(cigf_eval, np.ndarray):
            cigf_eval = np.asarray(cigf_eval)

        if not isinstance(cegf_eval, np.ndarray):
            cegf_eval = np.asarray(cegf_eval)

        # check if constraint functions are returning column vectors
        if cif_eval.ndim != 2:  # 2D
            ValueError("Inequality constraint function must return a 2D array. The current function is returning "
                       f"{cif_eval.ndim} dimension(s)")
        else:
            if cif_eval.shape[1] == 1 or cif_eval.shape[1] == 0:
                ValueError("Inequality constraint function must return a column array.")

        if cef_eval.ndim != 2:  # 2D
            ValueError("Equality constraint function must return a 2D array. The current function is returning "
                       f"{cef_eval.ndim} dimension(s)")
        else:
            if cef_eval.shape[1] == 1 or cef_eval.shape[1] == 0:
                ValueError("Inequality constraint function must return a column array.")

        # check if the jacobian is specified without its corresponding function evaluation
        if (cif_eval.size == 0 and cigf_eval.size != 0) or (cef_eval.size == 0 and cegf_eval.size != 0):
            ValueError("You can't specify the jacobian without their constraint functions")

        # check if the jacobians have their proper dimensions
        n_ce, n_var = cegf_eval.shape
        if n_ce != 0 and (n_ce != cef_eval.size or n_var != x0.size):
            ValueError("The jacobian of equality constraints must have its number of rows equal to the number "
                       "of equality constraint functions, and its number of columns equal to the number of elements "
                       "of the initial estimate.")

        n_ci, n_var = cigf_eval.shape
        if n_ci != 0 and (n_ci != cif_eval.size or n_var != x0.size):
            ValueError("The jacobian of inequality constraints must have its number of rows equal to the number "
                       "of inequality constraint functions, and its number of columns equal to the number of elements "
                       "of the initial estimate.")

    else:
        ValueError("Constraints function and their jacobian must be a callable function.")


# Empty non-linear constraint function (if nonlcon is empty)
def __empty_nonlcon():
    empty_eval = np.array([[]])
    return empty_eval, empty_eval, empty_eval, empty_eval


def __qp_solve(H, f, A=None, b=None, Aeq=None, beq=None, x0=None, solver=None):
    """
    Finds a minimum for a problem specified by
    min 0.5 * x^T * H * x + f^T * x
    subject to:
    A*x <= b
    Aeq*x = beq

    Parameters
    ----------
    H : numpy.array
        Symmetric array of doubles.
    f : numpy.array
        One-dimensional array of doubles.
    A : numpy.array
        Two-dimensional array of doubles. Linear coefficients of the inequality constraints.
    b : numpy.array
        One-dimensional array of doubles. Constant vector of the inequality constraints.
    Aeq : numpy.array
        Two-dimensional array of doubles. Linear coefficients of the equality constraints.
    beq : numpy.array
        One-dimensional array of doubles. Constant vector of the equality constraints.
    x0 : numpy.array
        One-dimensional array of doubles. Warm-start guess vector.
    solver: string
        Set to 'cvxopt' to run CVXOPT rather than quadprog.

    Returns
    -------
    x : numpy.array
        Solution that minimizes the quadratic problem.
    fval: double
        Value of the quadratic problem at the solution.
    exitflag: int
        Convergence flag of the problem. 1 for converged, 0 for failure.
    lambdav: dict
        Lagrange multipliers at the solution x. Contains the fields 'eq' and 'ineq' corresponding to the equality and
        inequality lagrange multipliers respectively.

    """
    # FIXME: (13/05/2019) CVXOPT behaving weirdly when using in caballero's algorithm, domain error. Changing to quadprog
    if solver == "cvxopt":  # CVXOPT
        H = matrix(H)
        f = matrix(f.reshape(-1).astype(np.double))

        if A.size == 0:
            A = None
            b = None
        else:
            A = matrix(A)
            b = matrix(b.reshape(-1).astype(np.double))

        if Aeq.size == 0:
            Aeq = None
            beq = None
        else:
            Aeq = matrix(Aeq)
            beq = matrix(beq.reshape(-1).astype(np.double))

        if x0.size == 0:
            x0 = None
        else:
            x0 = matrix(x0.reshape(-1).astype(np.double))

        if solver == "cvxopt" or solver is None:
            solver = None
        else:
            solver = "mosek"

        sol = cvxqp(H, f, A, b, Aeq, beq, solver=solver, initvals=None)

        x = np.array(sol['x']).flatten()
        fval = sol['primal objective']

        if sol['status'] == 'optimal':
            exitflag = 1
        else:
            exitflag = 0

        lambdav = {'eq': np.array(sol['y']).reshape(-1), 'ineq': np.array(sol['z']).reshape(-1)}

    elif solver == 'quadprog' or solver is None:
        meq = 0 if beq is None else beq.size

        if meq == 0:  # no equality constraints
            C_qp = -A.T
            b_qp = -b.flatten()
        else:
            C_qp = -np.vstack((Aeq, A)).T
            b_qp = -np.concatenate((beq.flatten(), b.flatten()))

        C_qp = None if C_qp.size == 0 else C_qp
        b_qp = None if b_qp.size == 0 else b_qp

        try:
            sol = solve_qp(H, -f.flatten(), C=C_qp, b=b_qp, meq=meq)
        except ValueError:  # QP failed
            x = x0 if x0 is not None else np.array([])
            fval = np.array([])
            exitflag = 0
            lambdav = {'eq': np.array([]), 'ineq': np.array([])}
        else:  # QP successful
            x, fval, _, _, lmbd, _ = sol
            exitflag = 1
            x = x.flatten()
            lmbd = lmbd.flatten()
            if meq != 0:
                active_set_idx = np.sort(sol[5] - 1)  # the -1 is to change from fortran index to c index
                lmbd_active_set = mldivide(C_qp[:, active_set_idx], H @ x + f.flatten())
            else:
                lmbd_active_set = np.array([])
            lambdav = {'eq': lmbd_active_set[:meq], 'ineq': lmbd[meq:]}

    return x, fval, exitflag, lambdav



import numpy as np
from scipy.optimize import root

from ..core.utils import _is_numeric_array_like


def evaporator(f1: float, f3: float, p100: float, f200: float,
               x1: float = 5.0, t1: float = 40.0, t200: float = 25.0,
               x0: list = None) -> dict:
    """Model function of an evaporation process.

    Parameters
    ----------
    f1 : float
        Feed flow rate value (kg/min).
    f3 : float
        Circulating flow rate (kg/min).
    p100 : float
        Steam pressure (kPa).
    f200 : float
        Cooling water flow rate (kg/min).
    x1 : float, optional
        Feed composition (molar %), by default 5.0.
    t1 : float, optional
        Feed temperature (C), by default 40.0.
    t200 : float, optional
        Inlet temperature of the cooling water (C), by default 25.0.
    x0 : list (floats), optional
        Initial estimate to be used by the non-linear solver. Must have 12
        elements.

    Returns
    -------
    results : dict
        Dictionary containing the results
    """

    def sub_model(x, params):
        f1, f3, p100, f200 = params

        f2, f4, f5, x2, t2, t3, p2, f100, t100, q100, t201, q200 = x

        quoc = (t3 - t200) / (0.14 * f200 + 6.84)

        eq = [
            (f1 - f4 - f2) / 20,
            (f1 * x1 - f2 * x2) / 20,
            (f4 - f5) / 4,
            0.5616 * p2 + 0.3126 * x2 + 48.43 - t2,
            0.507 * p2 + 55 - t3,
            (q100 - 0.07 * f1 * (t2 - t1)) / 38.5 - f4,
            0.1538 * p100 + 90 - t100,
            0.16 * (f1 + f3) * (t100 - t2) - q100,
            q100 / 36.6 - f100,
            0.9576 * f200 * quoc - q200,
            t200 + 13.68 * quoc - t201,
            q200 / 38.5 - f5
        ]

        return eq

    # initial estimate
    if x0 is None:
        x0 = (1000 * np.ones(12,)).tolist()
    else:
        # check dimension and type
        if _is_numeric_array_like(x0):
            x0 = np.asarray(x0, dtype=float)
        else:
            raise ValueError("'x0' has to be a float array.")

    # extra args
    param = (f1, f3, p100, f200)

    # solver call
    res = root(sub_model, x0, args=(param,))

    # unpack the results
    f2, f4, f5, x2, t2, t3, p2, f100, t100, q100, t201, q200 = \
        res['x'].tolist()

    # calculate the objective function
    j = 600 * f100 + 0.6 * f200 + 1.009 * (f2 + f3) + 0.2 * f1 - 4800 * f2

    # constraints
    g1 = 35.5 - x2
    g2 = p2 - 80
    g3 = 40 - p2

    g = [g1, g2, g3]

    # status of the run (True or False)
    status = res['success']

    # extra outputs (maybe useful/optional)
    extras = dict(zip(["f2", "f4", "f5", "x2", "t2", "t3", "p2", "f100",
                       "t100", "q100", "t201", "q200"], res['x'].tolist()))

    # results dictionary (must contain status, obj, const)
    results = {
        'status': status,
        'f': j,
        'g': g,
        'extras': extras
    }

    return results



import os
from string import Template

import numpy as np
from colorama import Fore, Style, deinit, init

from ..core.options import ProcedureOptions
from ..core.procedures.output import Report

__all__ = ["CaballeroOptions", "is_inside_hypercube", "CaballeroReport"]


class CaballeroReport(Report):
    """Report class specifically for caballero procedure only.
    """

    def __init__(self, terminal=False, plot=False):
        super().__init__(terminal=terminal, plot=plot)

    def build_iter_report(self, movement: str, iter_count: int, x: list,
                          f_pred: float, f_actual: float, g_actual: float,
                          header=False, field_size: int = 12):
        arr_str = super().build_iter_report(iter_count=iter_count, x=x,
                                            f_pred=f_pred, f_actual=f_actual,
                                            g_actual=g_actual, header=header,
                                            field_size=field_size)
        mv_temp = Template("{:$f_size}\t").substitute(f_size=field_size)

        # concatentate movement and iteration string report
        if header:
            mv_str = 'Last move'
        else:
            mv_str = movement

        return mv_temp.format(mv_str) + arr_str

    def print_iteration(self, movement: str, iter_count: int, x: list,
                        f_pred: float, f_actual: float, g_actual: float,
                        header=False, field_size: int = 12, color_font=None):

        arr_str = self.build_iter_report(movement=movement,
                                         iter_count=iter_count, x=x,
                                         f_pred=f_pred, f_actual=f_actual,
                                         g_actual=g_actual, header=header,
                                         field_size=field_size)

        if self.terminal:
            # terminal asked, check font color
            if color_font == 'red':
                print(Fore.RED + arr_str)
            else:
                print(Fore.RESET + arr_str)


class CaballeroOptions(ProcedureOptions):
    """Options structure for the Caballero class algorithm.

    Parameters
    ----------
    max_fun_evals : int, optional
        Maximum number of black box function evaluations, by default 500.

    feasible_tol : float, optional
        Feasibility tolerance for the model constraints g(x) <= tol, by default
        1e-06.

    penalty_factor : float, optional
        Value to penalize the objective function when its value returned by the
        black box model indicates that is a infeasible result (note that by
        infeasible it is referring to whether or not the sampling converged for
        that case, not optimization feasibility), by default None.

        See this property notes for more info.

    ref_tol : float, optional
        Refinement tolerance specification (tol1), by default 1e-4.

    term_tol : float, optional
        Termination tolerance specification (tol2), by default 1e-5. Has to be
        lesser than `ref_tol`.

    first_factor : float, optional
        First contraction factor specification, by default 0.6. Has to be
        between 0 and 1.

    second_factor : float, optional
        Subsequent contraction factor specification, by default 0.4. Has to be
        between 0 and 1. Has to be lesser than `first_factor`.

    contraction_tol : float, optional
        Maximum contraction size that the refinement hypercube can achieve
        when compared to the original domain, by default 1e-4.

        See this property notes for more information.
    """
    @property
    def penalty_factor(self):
        """Value to penalize the objective function when its value returned by
        the black box model indicates that is a infeasible result (note that
        "infeasible result" it is referring to whether or not the sampling
        converged for that case, not constraint feasibility), by default None.

        The value set for the penalty factor is simply summed to the objective
        function value in order to make the optimization procedure avoid
        sampling in the known infeasible regionself.

        If this value is set to None, the procedure will automatically chose
        the highest value of the objective function in the initial sampling as
        the penalty factor."""
        return self._penalty_factor

    @penalty_factor.setter
    def penalty_factor(self, value):
        self._penalty_factor = value

    @property
    def first_factor(self):
        """First contraction factor specification."""
        return self._first_factor

    @first_factor.setter
    def first_factor(self, value):
        if isinstance(value, float):
            if 0.0 < value < 1.0:
                self._first_factor = value
            else:
                raise ValueError("'first_factor' value has to be in the "
                                 "range (0, 1).")

        else:
            raise ValueError("'first_factor' has to be a float.")

    @property
    def second_factor(self):
        """Second contraction factor specification."""
        return self._second_factor

    @second_factor.setter
    def second_factor(self, value):
        if isinstance(value, float):
            if 0.0 < value < 1.0:
                self._second_factor = value
            else:
                raise ValueError("'second_factor' value has to be in the "
                                 "range (0, 1).")

        else:
            raise ValueError("'second_factor' has to be a float.")

    @property
    def ref_tol(self):
        """Refinement tolerance specification (tol1)."""
        return self._ref_tol

    @ref_tol.setter
    def ref_tol(self, value):
        if isinstance(value, float):
            if value > 0.0:
                self._ref_tol = value
            else:
                raise ValueError("'ref_tol' value has to be a positive float.")

        else:
            raise ValueError("'ref_tol' has to be a float.")

    @property
    def term_tol(self):
        """Termination tolerance specification (tol2)."""
        return self._term_tol

    @term_tol.setter
    def term_tol(self, value):
        if isinstance(value, float):
            if value > 0.0:
                self._term_tol = value
            else:
                raise ValueError("'term_tol' value has to be a positive "
                                 "float.")

        else:
            raise ValueError("'term_tol' has to be a float.")

    @property
    def contraction_tol(self):
        """Maximum contraction size that the refinement hypercube can achieve
        when compared to the original domain.

        The ratio between the current refined hypercube range and the original
        domain range can't be lesser than `contraction_tol`. After the ratio
        reaches this value, no more contractions will be perfomed."""
        return self._contraction_tol

    @contraction_tol.setter
    def contraction_tol(self, value):
        if isinstance(value, float):
            if value > 0.0:
                self._contraction_tol = value
            else:
                raise ValueError("'contraction_tol' value has to be a positive"
                                 " float.")

        else:
            raise ValueError("'contraction_tol' has to be a float.")

    # -------------------------------------------------------------------------
    def __init__(self, max_fun_evals: int = 500, feasible_tol: float = 1e-06,
                 penalty_factor: float = None,
                 ref_tol: float = 1e-4, term_tol: float = 1e-5,
                 first_factor: float = 0.6, second_factor: float = 0.4,
                 contraction_tol: float = 1e-4):

        super().__init__(max_fun_evals=max_fun_evals,
                         feasible_tol=feasible_tol)

        self.penalty_factor = penalty_factor
        self.first_factor = first_factor
        self.second_factor = second_factor
        self.ref_tol = ref_tol
        self.term_tol = term_tol
        self.contraction_tol = contraction_tol

        # perform instance checkup
        self.check_options_setup()

    def check_options_setup(self):
        if self.first_factor <= self.second_factor:
            raise ValueError("'first_factor' has to be greater than "
                             "'second_factor'.")

        if self.ref_tol <= self.term_tol:
            raise ValueError("'ref_tol' has to be greater than "
                             "'term_tol'.")


def is_inside_hypercube(point: np.ndarray, lb: np.ndarray, ub: np.ndarray,
                        tol: float = 1e-8):
    """Determines if a `point` is inside a hypercube following a specified
    tolerance.

    Parameters
    ----------
    point : np.ndarray
        Point to be checked (1D array).

    lb : np.ndarray
        Hypercube lower bound (1D array).

    ub : np.ndarray
        Hypercube upper bound(1D array).

    tol : float, optional
        Tolerance value to accept whether or not `point` is at the limit.
        Default is 1e-8. (This value is a percentage of domain range.)
    """

    # Offset the hypercube to the origin
    lbf = np.zeros(lb.shape)
    ubf = ub - lb

    # offset the point in relation to the origin as well
    pointf = point - lb

    # check if the point is outside the hypercube. If so, raise exception
    if np.any(np.greater(pointf, ubf)) or np.any(np.less(pointf, lbf)):

        def arr2str_fcn(x): return np.array2string(x, precision=4,
                                                   separator='\t', sign=' ')

        err_msg = ("Point: {0}\nLB: {1}\nUB: {2}".format(arr2str_fcn(point),
                                                         arr2str_fcn(lb),
                                                         arr2str_fcn(ub)))
        raise ValueError("The point is outside the hypercube.\n" + err_msg)
    else:
        # check if the point is inside or at limit
        if np.any(np.abs((ubf - pointf) / ubf) <= tol) \
                or np.any(np.abs((pointf - lbf) / ubf <= tol)):
            # limit
            return False
        else:
            # inside
            return True



import numpy as np
from scipy.spatial.distance import pdist


def _is_numeric_array_like(value):
    """Returns True if the input is a valid float array.
    """
    if isinstance(value, np.ndarray):
        return True if value.dtype == float else False

    else:
        # value is not array, try to convert it to numpy float array
        try:
            value = np.asarray(value, dtype=float)
        except ValueError:
            # conversion failed, input can't be converted to float array
            return False
        finally:
            # conversion successful
            return True


def is_row_member(a: np.ndarray, b: np.ndarray):
    """If `a` is a 1D array, checks if `a` is a row of `b`. Otherwise, checks
    if each row of `a` is row of b (the result is a 1D bool array).

    Parameters
    ----------
    a : np.ndarray
        1D or 2D array to be checked. If 1D, treats `a` as single row.
    b : np.ndarray
        2D array where `a` is checked against.

    Returns
    -------
    out : bool or 1D bool array
        If `a` is 1D, the result is a single boolean value, else a 1D bool
        array where each element corresponds to a row of `a`.
    """
    if b.ndim != 2:
        raise ValueError("b has to be a 2D array.")

    if a.ndim == 1:
        # a is a 1D array, treat the entire array as row
        return (a == b).all(axis=1).any()
    elif a.ndim == 2:
        # a is a 2D array, return a bool 1D array
        a_rows, _ = a.shape
        result = np.full((a_rows,), False, dtype=bool)

        for row in range(a_rows):
            result[row] = is_row_member(a[row], b)

        return result

    else:
        raise ValueError("Row check valid only for 1D or 2D arrays.")


def point_domain_distance(p1: np.ndarray, p2: np.ndarray, lb: np.ndarray,
                          ub: np.ndarray) -> float:
    """Computes the euclidian distance between two points (`p1`, `p2`) inside a
    box domain defined by a lower bound (`lb`) and an upper bound (`ub`). The
    result is the normalized distance between `p1` and `p2` (i.e. the
    percentage of the domain range which the distance corresponds).

    Parameters
    ----------
    p1 : np.ndarray
        Point 1 (1D array).

    p2 : np.ndarray
        Point 2 (1D array).

    lb : np.ndarray
        Domain lower bound (1D array).

    ub : np.ndarray
        Domain upper bound (2D array).

    Returns
    -------
    dist : float
        Normalized euclidian distance.
    """

    if p1.ndim != 1 or p2.ndim != 1 or lb.ndim != 1 or ub.ndim != 1:
        raise ValueError("All input points must be 1D arrays.")

    return (pdist(np.vstack((p1, p2)), metric='euclidean') /
            pdist(np.vstack((lb, ub)), metric='euclidean')).item()


def get_samples_index(x: np.ndarray, lb: np.ndarray, ub: np.ndarray):
    """Get the indexes of `x` that are inside the hypercube bounded by `lb` and
    `ub`.

    Parameters
    ----------
    x : np.ndarray
        Sample set (2D array).
    lb : np.ndarray
        Lower bound of hypercube (1D array).
    ub : np.ndarray
        Upper bound of hypercube (1D array).

    Returns
    -------
    out : np.ndarray
        Indexes (row numbers) of `x` that are between `lb` and `ub`.
    """
    idx = np.nonzero(np.all(np.logical_and(np.less_equal(x, ub),
                                           np.greater_equal(x, lb)),
                            axis=1))[0]

    return idx


class CaballeroProblem(object):
    """Class problem interface for IpOpt use."""

    def __init__(self, obj_surrogate: dict, con_surrogate: list):
        self._obj_surr = obj_surrogate
        self._con_surr = con_surrogate

    def objective(self, x):
        f, *_ = self._obj_surr.predict(x)

        return f

    def gradient(self, x):
        _, gf, *_ = self._obj_surr.predict(x, compute_jacobian=True)

        return gf

    def constraints(self, x):
        c = np.zeros((len(self._con_surr), 1))

        for i in range(len(self._con_surr)):
            c[i, 0], *_ = self._con_surr[i].predict(x)

        return c

    def jacobian(self, x):
        gc = np.zeros((len(self._con_surr), x.size))

        for i in range(len(self._con_surr)):
            _, gc_col, *_ = self._con_surr[i].predict(x, compute_jacobian=True)
            gc[[i], :] = gc_col.reshape(1, -1)

        return gc



import json
from abc import ABC, abstractmethod

import ipopt
import numpy as np
import requests
from pydace import Dace

from ..options.nlp import DockerNLPOptions, IpOptOptions, NLPOptions
from ..procedures import InfillProcedure
from ..utils import CaballeroProblem


def optimize_nlp(procedure: InfillProcedure, x: np.ndarray, g: np.ndarray,
                 f: np.ndarray, nlp_options: NLPOptions, x0: list, lb: list,
                 ub: list):
    if not isinstance(procedure, InfillProcedure):
        raise ValueError("'procedure' has to be a valid Infill procedure "
                         "object.")

    if not isinstance(nlp_options, NLPOptions):
        raise ValueError("'nlp_options' has to be a valid Non-Linear options "
                         "object.")

    if isinstance(nlp_options, DockerNLPOptions):
        # use docker server

        # constraint hyperparameters
        con_theta = [model.theta.flatten().tolist()
                     for model in procedure.surr_con]

        # process data to be sent
        surr_data = {'input_design': x.tolist(),
                     'fobj_data': {
            'fobj_obs': f.tolist(),
            'fobj_theta': procedure.surr_obj.theta.flatten().tolist()},
            'const_data': {
            'const_obs': g.tolist(),
            'const_theta': con_theta},
            'regmodel': procedure.regression,
            'corrmodel': 'corrgauss'
        }

        # ipopt options
        nlp_opts = {
            'tol': nlp_options.tol,
            'max_iter': nlp_options.max_iter,
            'con_tol': nlp_options.con_tol
        }

        dic = {'x0': x0,
               'lb': lb,
               'ub': ub,
               'surr_data': surr_data,
               'nlp_opts': nlp_opts}

        # send data to server and return the nlp solution
        response = requests.post(url=nlp_options.server_url + '/opt', json=dic)
        sol = response.json()
        sol['x'] = np.array(sol['x'])  # converting from list to np.array

    elif isinstance(nlp_options, IpOptOptions):
        # local installation of IpOpt solver
        x0 = np.array(x0).flatten()
        lb = np.array(lb).flatten()
        ub = np.array(ub).flatten()

        m, n = x.shape

        # hyperparameters
        f_theta = procedure.surr_obj.theta.flatten()
        con_theta = [model.theta.flatten() for model in procedure.surr_con]

        # build the surrogates
        obj_surr = Dace(regression=procedure.regression,
                        correlation='corrgauss')
        obj_surr.fit(S=x, Y=f, theta0=f_theta)

        con_surr = []
        for j in range(g.shape[1]):
            con_surr_ph = Dace(regression=procedure.regression,
                               correlation='corrgauss')
            con_surr_ph.fit(S=x, Y=g[:, j],
                            theta0=con_theta[j])
            con_surr.append(con_surr_ph)

        # ------------------------------ Solver call ------------------------------
        nlp = ipopt.problem(
            n=x0.size,
            m=len(con_surr),
            problem_obj=CaballeroProblem(obj_surr, con_surr),
            lb=lb,
            ub=ub,
            cl=-np.inf * np.ones(len(con_surr)),
            cu=np.zeros(len(con_surr))
        )

        # ipopt options
        nlp.addOption('tol', nlp_options.tol)
        nlp.addOption('constr_viol_tol', nlp_options.con_tol)
        nlp.addOption('max_iter', nlp_options.max_iter)
        nlp.addOption('hessian_approximation', 'limited-memory')
        nlp.addOption('print_level', 0)
        nlp.addOption('mu_strategy', 'adaptive')

        x, info = nlp.solve(x0)
        fval = info['obj_val']
        exitflag = info['status']

        if exitflag == 0 or exitflag == 1 or exitflag == 6:
            # ipopt succeded
            exitflag = 1
        else:
            # ipopt failed. See IpReturnCodes_inc.h for complete list of flags
            exitflag = 0

        return {'x': x, 'fval': fval, 'exitflag': exitflag}

    else:
        raise NotImplementedError("NLP solver not implemented.")

    return sol



from abc import ABC, abstractmethod


class ProcedureOptions(ABC):
    """Base class for infill procedure options configuration."""

    _VALID_NLP_SOLVERS = ['ipopt']

    @property
    def max_fun_evals(self):
        """Maximum number of (rigorous) model function evaluations."""
        return self._max_fun_evals

    @max_fun_evals.setter
    def max_fun_evals(self, value):
        if isinstance(value, int):
            if value > 0:
                self._max_fun_evals = value
            else:
                raise ValueError("'max_fun_evals' has to be a positive "
                                 "integer.")

        else:
            raise ValueError("'max_fun_evals' has to be a integer.")

    @property
    def feasible_tol(self):
        """Feasibility tolerance for the model constraints g(x) <= tol."""
        return self._feasible_tol

    @feasible_tol.setter
    def feasible_tol(self, value):
        if isinstance(value, float):
            if value > 0.0:
                self._feasible_tol = value
            else:
                raise ValueError("'feasible_tol' has to be a positive small "
                                 "value.")

        else:
            raise ("'feasible_tol' has to be a float.")

    # -------------------------------------------------------------------------

    def __init__(self, max_fun_evals: int = 500, feasible_tol: float = 1e-6):
        # initialize base class
        super().__init__()

        # class base parameters
        self.max_fun_evals = max_fun_evals
        self.feasible_tol = feasible_tol

    @abstractmethod
    def check_options_setup(self):
        """Abstract method to be overriden where the procedure options are
        checked before the optimization starts.
        """
        pass



import logging

import numpy as np

from surropt.optimizers.utils import (__bnd2cf, __check_vector_input,
                                      __constraint_function_check,
                                      __empty_nonlcon, __linesearch,
                                      __objective_function_check, __qp_solve,
                                      __set_options_structure)
from surropt.utils.matrixdivide import mrdivide
from tests_ import OPTIMIZERS_PATH


# module variables
LOG_OFF = False  # flag to turn on loggin of iterations

# logger configuration
LOG_LEVEL = logging.DEBUG
LOG_FORMAT = "%(levelname)s - %(message)s"
logging.basicConfig(filename=OPTIMIZERS_PATH / "sqp_log.log",
                    level=LOG_LEVEL,
                    format=LOG_FORMAT,
                    filemode='w')
logger = logging.getLogger()


def sqp(objfun: callable, x0: np.ndarray, confun: callable = None, lb=None,
        ub=None, options: dict = None):
    """
    This SQP implementation is the one described by [1]_ and implemented in
    Octave optimization toolbox.

    Min         f(x)
    subject to:
                c_i(x) <= 0,  for i = 1, ..., n

    where f(x) and c_i(x) are the objective and constraint functions

    Parameters
    ----------
    objfun : callable
        Objective function. Its gradient must be evaluated inside this function
        and returned as 2-D column array. See notes.

    x0 : numpy.array
        Initial estimate. For now, this implementation of SQP doesn't handle
        initial infeasibility. Therefore, `x0` hasto be feasible.

    confun : callable
        Constraints and their Jacobian evaluation, both inequality and
        equality. See notes.

    lb : {None, numpy.array}
        Lower bounds. Same number of elements as `x0`. If any of the bounds
        arguments are set as None (default), it means that all elements of that
        bound are set as positive(upper bound)/negative(lower bound) infinity
        1-D array.

    ub : {None, numpy.array}
        Upper bounds. If any of the bounds arguments are set as None (default),
        it means that all elements of that bound are set as positive(upper
        bound)/negative(lower bound) infinity 1-D array.

    options : dict
        Dictionary containing fields with algorithm options. Valid options are:

            - maxiter : maximum number of iterations. Default is 400;
            - maxfunevals : maximum function evaluations. Default is
                            100 * `x0`.size;
            - tolopt : Optimality tolerance. Default is 1e-6;
            - tolcon : Constraint tolerance. Default is 1e-6;
            - tolstep : Minimum step size tolerance. Default is 1e-6;
            - qpsolver: Which library to use the quadratic programming (QP)
                        solver. Default is 'quadprog'.

    Returns
    -------
    sol : dict
        Solution dictionary containing the fields:
        'x' : numpy.array
            Local solution if `exitflag` is positive.
        'fval' : double
            Objective function value at the solution `x`.
        'exitflag' : int
            Convergence flag. Values are:
            - 1 : First-order optimality measure is less than
                 `options`['tolopt'] and minimum constraint violation was less
                  than `options`['tolcon'].
            - 2 : Step size is less than `options`['tolstep'] and minimum
                  constraint violation was less than `options`['tolcon'].
            - 0 : Maximum number of iterations achieved.
            - -1 : BFGS update failed.
            - -2 : Step size is less than `options`['tolstep']. However there
                   still is some significant constraint violation.
        'nfevals' : int
            Number of objective function evaluation.
        'lambda' : dict
            Dictionary containing equality ('eq'), inequality ('ineq') and
            bounds ('lb', 'ub') Lagrange multipliers at the solution `x`.
        'iterations' : int
            Number of iterations performed by the algorithm.

    Raises
    ------
    ValueError
        If `x0`,`lb` or `ub` aren't numeric arrays with same number of elements
        If `lb` is greater `ub` for any element.
        If an invalid option parameter is set, e.g. typo in name parameter or
        the parameter doesn't exist.
        If the `objfun` or `confun` was set improperly.

    Notes
    -----
        ** How to set the `objfun` function **

        The `objfun` object must return a tuple of two elements. The first
        being the objective function evaluation (float) and the second being a
        2-D column array containing the gradient of objective function.

        ** How to set the `confun` function **
        The `confun` callable object must return a tuple of four elements. The
        first two elements are the evaluations of inequality and equality
        constraints, respectively, returned as 2-D column array. The third and
        fourth elements of this tuple are the Jacobian evaluation of equality
        and inequality constraints, respectively. If the problem has only
        equalities, you must set the first and third elements as 2-D empty
        array (numpy.array([[]]). If it only has inequalities, set the second
        and fourth elements as 2-D empty array.


    References
    ----------
    .. [1] Nocedal, J. and S. J. Wright. Numerical Optimization, Second
        Edition. Springer Series in Operations Research, Springer Verlag, 2006.

    Examples
    --------
    ** Single equality constraint, single inequality constraint and 4 bounds **

    Define the objective function as:

    >>> import numpy as np
    >>> def objective(x):
    ... # objective function - HS071 test problem
    ... obj = x[0]*x[3]*(x[0]+x[1]+x[2])+x[2]
    ...
    ... # gradient
    ... obj_grad = np.zeros((4,))
    ... obj_grad[0] = x[0] * x[3] + x[3]*(x[0]+x[1]+x[2])
    ... obj_grad[1] = x[0] * x[3]
    ... obj_grad[2] = x[0] * x[3] + 1
    ... obj_grad[3] = x[0]*(x[0]+x[1]+x[2])
    ...
    ... return obj, obj_grad

    Constraints:

    >>> def constraints(x):
    ...
    ... # define the inequality constraints
    ... ineq = np.array([[25 - x[0]*x[1]*x[2]*x[3]]])
    ... # define the inequality constraints Jacobian
    ... ineq_grad = np.zeros((1, 4))
    ... ineq_grad[0, 0] = -x[1] * x[2] * x[3]
    ... ineq_grad[0, 1] = -x[0] * x[2] * x[3]
    ... ineq_grad[0, 2] = -x[0] * x[1] * x[3]
    ... ineq_grad[0, 3] = -x[0] * x[1] * x[2]
    ...
    ... # define the equality constraints
    ... eq = np.array([[np.sum(np.multiply(x, x)) - 40]])
    ... # define the equality constraints Jacobian
    ... eq_grad = np.zeros((1, 4))
    ... eq_grad[0, 0] = 2 * x[0]
    ... eq_grad[0, 1] = 2 * x[1]
    ... eq_grad[0, 2] = 2 * x[2]
    ... eq_grad[0, 3] = 2 * x[3]
    ...
    ... return ineq, eq, ineq_grad, eq_grad

    Initial estimate

    >>> x0 = np.array([1., 5., 5., 1.])

    Call the solver

    >>> sol = sqp(objective, x0, constraints)
    >>> x = sol['x']
    >>> fval = sol['fval']
    >>> exitflag = sol['exitflag']
    >>> nfunevals = sol['nfevals']
    >>> lagrange_mult = sol['lambda']
    >>> iterations = sol['iterations']

    The values are:

    - x = [1. 4.74299964 3.82114998 1.37940829]
    - fval = 17.014017289178593
    - exitflag = 2
    - nfunevals = 7
    - lagrange_mult['eq'] = [-0.16146857]
    - lagrange_mult['ineq'] = [-0.55229366]
    - lagrange_mult['lb'] = [-1.08787123e+00 -3.40273704e-12 -8.67542702e-12 -5.52639122e-12]
    - lagrange_mult['ub'] = [-3.46930334e-12 -6.12161682e-10 -8.34873225e-12 -3.39974223e-12]
    - iterations = 5


    ** Three equality constraints, no inequalities and no bounds problem **

    Define the objective function as:

    >>> import numpy as np
    >>> def objective(x):
    ... # objective function
    ... obj = np.exp(np.prod(x)) - 0.5*(x[0]**3 + x[1]**3 + 1)**2
    ...
    ... # gradient
    ... gradobj = np.zeros((5, 1))
    ... gradobj[0] = (np.prod(x[1:]) * np.exp(np.prod(x)) - (3*x[0]**2)*(x[0]**3+x[1]**3+1))
    ... gradobj[1] = np.prod(x[np.arange(len(x)) != 1]) * np.exp(np.prod(x)) - (3*x[1]**2)*(x[0]**3+x[1]**3+1)
    ... gradobj[2] = np.prod(x[np.arange(len(x)) != 2]) * np.exp(np.prod(x))
    ... gradobj[3] = np.prod(x[np.arange(len(x)) != 3]) * np.exp(np.prod(x))
    ... gradobj[4] = np.prod(x[np.arange(len(x)) != 4]) * np.exp(np.prod(x))
    ...
    ... return obj, gradobj

    Constraints:

    >>> def constraints(x):
    ...
    ... # define the equality constraints
    ... r = np.array([
    ...     [np.sum(x**2) - 10],
    ...     [x[1]*x[2] - 5*x[3]*x[4]],
    ...     [x[0]**3 + x[1]**3 + 1]
    ... ])
    ... # define the equality constraints Jacobian
    ... rgrad = np.zeros((3, 5))
    ... rgrad[0, 0] = 2 * x[0]
    ... rgrad[0, 1] = 2 * x[1]
    ... rgrad[0, 2] = 2 * x[2]
    ... rgrad[0, 3] = 2 * x[3]
    ... rgrad[0, 4] = 2 * x[4]
    ...
    ... rgrad[1, 0] = 0
    ... rgrad[1, 1] = x[2]
    ... rgrad[1, 2] = x[1]
    ... rgrad[1, 3] = -5 * x[4]
    ... rgrad[1, 4] = -5 * x[3]
    ...
    ... rgrad[2, 0] = 3 * x[0] ** 2
    ... rgrad[2, 1] = 3 * x[1] ** 2
    ... rgrad[2, 2] = 0
    ... rgrad[2, 3] = 0
    ... rgrad[2, 4] = 0
    ...
    ... return np.array([[]]), r, np.array([[]]), rgrad

    Initial estimate

    >>> x0 = np.array([-1.8, 1.7, 1.9, -0.8, -0.8])

    Call the solver

    >>> sol = sqp(objective, x0, constraints)
    >>> x = sol['x']
    >>> fval = sol['fval']
    >>> exitflag = sol['exitflag']
    >>> nfunevals = sol['nfevals']
    >>> lagrange_mult = sol['lambda']
    >>> iterations = sol['iterations']

    The values are:

    - x = [-1.71714355  1.59570967  1.82724579 -0.76364308 -0.76364308]
    - fval = 0.05394984777028391
    - exitflag = 2
    - nfunevals = 9
    - lagrange_mult['eq'] = [-0.04016274  0.03795777 -0.00522264]
    - lagrange_mult['ineq'] = []
    - lagrange_mult['lb'] = []
    - lagrange_mult['ub'] = []
    - iterations = 7


    """

    if LOG_OFF:
        logger.disabled = True

    # check input vectors x0, lb, ub
    x0, lb, ub = __check_vector_input(x0, lb, ub)

    logger.debug("Initial estimate (x0) - " +
                 np.array2string(x0, precision=4, separator=',',
                                 suppress_small=True))
    logger.debug("Iter #\t|\tx_i\t|\tp_i\t|\tfval")

    # check the objective function
    __objective_function_check(objfun, x0)

    # check options dictionary
    maxiter, maxfunevals, tolstep, tolopt, tolcon, qpsolver = \
        __set_options_structure(options, x0)

    # check if any confun parameters is coming as None (withtout constraints)
    # and make it a empty handle
    if confun is None:
        confun = __empty_nonlcon()

    # check constraints
    __constraint_function_check(confun, x0)

    lbgrad = - np.eye(x0.size)
    ubgrad = - lbgrad

    lbidx = lb != - np.inf
    ubidx = ub != np.inf
    lbgrad = lbgrad[lbidx, :]
    ubgrad = ubgrad[ubidx, :]

    # Transform bounds into inequality constraints
    def confun(xv, fun=confun): return __bnd2cf(xv, lbidx, ubidx, lb, ub,
                                                np.vstack((lbgrad, ubgrad)),
                                                fun)

    # global structure for parametrization
    globalls = {'nevals': 0}

    # Initialize variables: objective and constraints
    x = x0

    obj, c = objfun(x0)
    globalls['nevals'] = 1

    # Initialize the positive definite Hessian Approximation
    n = x0.size
    B = np.eye(n)

    # Evaluate the constraint info
    ci, ce, C, F = confun(x0)

    # Jacobian Matrix of bothe equalites and inequalites
    A = np.vstack((F, C))

    # Choose an initial lambda

    lambdav = 100 * np.ones((ce.size + ci.size,))

    iteration = 0

    while iteration < maxiter:
        # Convergence check
        nr_f = ce.size  # Number of equality constraints

        # Split lagrange multipliers
        lambda_i = lambdav[nr_f:]

        con = np.vstack((ce, ci))

        # Karush-Kuhn-Tucker conditions

        t0 = np.linalg.norm(c - A.conj().T @ lambdav)
        t1 = np.linalg.norm(ce)
        t2 = np.all(ci <= 0)
        t3 = np.all(lambda_i >= 0)
        t4 = np.linalg.norm(lambdav * con)

        if t2 and t3 and np.max([t0, t1, t4]) < tolopt:
            # Problem has converged with all constraints being satisfied
            exitflag = 1
            break

        # Solve the QP subproblem to compute the search direction p
        lambda_old = lambdav.copy()  # Store old multipliers

        p, qpfval, qpexflag, lambdadict = __qp_solve(
            B, c, C, -ci, F, -ce, x0=np.array([]), solver=qpsolver)

        if qpexflag == 1:
            lambdav[:nr_f] = -lambdadict['eq']
            lambdav[nr_f:] = -lambdadict['ineq']
        else:
            lambdav = lambda_old.copy()
            p = x

        # Perform linesearch
        x_new, alpha, obj_new, c_new, globalls = __linesearch(
            x, p, objfun, confun, lambdav, obj, c, globalls)

        # Re-evaluate objective, constraints and gradients at new x value
        ci_new, ce_new, C_new, F_new = confun(x_new)

        A_new = np.vstack((F_new, C_new))

        y = c_new - c

        if A.size != 0:
            t = (A_new - A).T @ lambdav
            y -= t

        delx = x_new - x  # Step size

        # Check if step size is too small
        if np.linalg.norm(delx) < tolstep * np.linalg.norm(x):
            # Check for minimum constraint violation
            if np.vstack((ce_new, ci_new)).size == 0 or \
                    np.max(np.vstack((ce_new, ci_new))) < tolcon:
                # the first verification is for cases when there are no
                # constraints, so there is no exception when using the max

                # Step size is too small and constraint violation is less than
                # minimum constraint violation
                exitflag = 2
            else:
                # Step size is too small but there some significant constraint
                # violation
                exitflag = -2

            break

        # Check for number of function evaluations
        if globalls['nevals'] > maxfunevals:
            exitflag = 0
            break

        delxt = delx[:, np.newaxis].conj().T

        # the item is to guarantee that the result is a scalar
        d1 = (delxt @ B @ delx).item()

        t1 = 0.2 * d1
        t2 = delxt @ y

        if t2 < t1:
            theta = np.asscalar(0.8 * d1 / (d1 - t2))
        else:
            theta = 1

        r = theta * y + (1 - theta) * B @ delx

        d2 = (delxt @ r).item()

        if d1 == 0 or d2 == 0:
            exitflag = -1  # BFGS Update failed
            break

        # the new axis is to regularize the algebraic vector operations
        # (r and delx are column vectors)
        B = B - mrdivide(B @ delx[:, np.newaxis] @ delxt @ B, d1) + \
            mrdivide(r[:, np.newaxis] @ r[:, np.newaxis].conj().T, d2)

        str_fmt = "{0} |\t{1} |\t{2} |\t{3}"
        x_str = np.array2string(x.flatten(), precision=4,
                                separator=',', suppress_small=True)
        p_str = np.array2string(p.flatten(), precision=4,
                                separator=',', suppress_small=True)
        logger.debug(str_fmt.format(iteration, x_str, p_str, obj))

        # Update new values
        x = x_new
        obj = obj_new
        c = c_new
        ce = ce_new
        F = F_new
        ci = ci_new
        C = C_new
        A = A_new

        iteration += 1

    if iteration >= maxiter:
        exitflag = 0  # Too many iterations

    nevals = globalls['nevals']

    n_c_eq = ce.size
    n_bnds = 0 if np.all(lb == -np.inf) and np.all(ub == np.inf) else lb.size
    n_c_iq = ci.size - 2 * n_bnds
    lmbda_dict = {'eq': lambdav[:n_c_eq],
                  'ineq': lambdav[n_c_eq:(n_c_eq + n_c_iq)],
                  'lb': lambdav[(n_c_eq + n_c_iq):(n_c_eq + n_c_iq + n_bnds)],
                  'ub': lambdav[(n_c_eq + n_c_iq + n_bnds): (n_c_eq + n_c_iq + 2 * n_bnds)]}

    return {'x': x, 'fval': obj, 'exitflag': exitflag, 'lambda': lmbda_dict,
            'iterations': iteration, 'nfevals': nevals}
    # end sqp



import requests
from requests.exceptions import ConnectionError
from abc import ABC


class NLPOptions(ABC):
    """Base (abstract) class for setting the NLP solver options.

    All the NLP solvers settings have to be handled through this class and 
    interfaced in the `optimize_nlp` function.
    """
    @property
    def name(self):
        """The solver name."""
        return self._name

    @name.setter
    def name(self, value):
        if isinstance(value, str):
            self._name = value
        else:
            raise ValueError("'name' has to be a string.")

    # -------------------------------------------------------------------------

    def __init__(self, name: str):
        self.name = name


class IpOptOptions(NLPOptions):
    """Local installation options for IpOpt solver.

    tol : float
        Desired convergence tolerance (relative). Determines the convergence
        tolerance for the algorithm. The algorithm terminates successfully, if
        the (scaled) NLP error becomes smaller than this value. Default is 1e-8

    max_iter : int
        Maximum number of iterations. The algorithm terminates with an error
        message if the number of iterations exceeded this number. The
        valid range for this integer option is 0 < max iter < +inf and its
        default value is 3000.

    con_tol : int
        Desired threshold for the constraint violation. Absolute tolerance on
        the constraint violation. Successful termination requires that the
        max-norm of the (unscaled) constraint violation is less than this
        threshold. The valid range for this real option is
        0 < constr viol tol < +inf and its default value is 1e-4.
    """
    @property
    def tol(self):
        """Desired convergence tolerance (relative)."""
        return self._tol

    @tol.setter
    def tol(self, value):
        if isinstance(value, float):
            if value > 0:
                self._tol = value
            else:
                raise ValueError("'tol' can't be zero or negative value.")
        else:
            raise ValueError("'tol' has to be a float.")

    @property
    def max_iter(self):
        """Maximum number of iterations."""
        return self._max_iter

    @max_iter.setter
    def max_iter(self, value):
        if isinstance(value, int):
            if value > 0:
                self._max_iter = value
            else:
                raise ValueError("'max_iter' can't be zero or negative value.")
        else:
            raise ValueError("'max_iter' has to be a integer.")

    @property
    def con_tol(self):
        """Desired threshold for the constraint violation."""
        return self._con_tol

    @con_tol.setter
    def con_tol(self, value):
        if isinstance(value, float):
            if value > 0:
                self._con_tol = value
            else:
                raise ValueError("'con_tol' can't be zero or negative value.")
        else:
            raise ValueError("'con_tol' has to be a float.")
    # -------------------------------------------------------------------------

    def __init__(self, name: str, tol: float = 1e-8, max_iter: int = 3000,
                 con_tol: float = 1e-4):
        super().__init__(name)

        self.tol = tol
        self.max_iter = max_iter
        self.con_tol = con_tol


class DockerNLPOptions(NLPOptions):
    """Solver options for when using IpOpt solver that is interfaced with a
    flask application inside a Docker container or WSL enviroment.

    Parameters
    ----------
    name : str
        Custom name to the `NLPOptions` object (i.e. just a identifier, not
        used as check anywhere else.)

    server_url : str
        Ip address of the docker server.

    tol : float
        Desired convergence tolerance (relative). Determines the convergence
        tolerance for the algorithm. The algorithm terminates successfully, if
        the (scaled) NLP error becomes smaller than this value. Default is 1e-8

    max_iter : int
        Maximum number of iterations. The algorithm terminates with an error
        message if the number of iterations exceeded this number. The
        valid range for this integer option is 0 < max iter < +inf and its
        default value is 3000.

    con_tol : int
        Desired threshold for the constraint violation. Absolute tolerance on
        the constraint violation. Successful termination requires that the
        max-norm of the (unscaled) constraint violation is less than this
        threshold. The valid range for this real option is
        0 < constr viol tol < +inf and its default value is 1e-4.
    """
    @property
    def server_url(self):
        """Ip address of the docker server."""
        return self._server_url

    @server_url.setter
    def server_url(self, value):
        if isinstance(value, str):
            self._server_url = value
        else:
            raise ValueError("'server_url' has to be a string.")

    @property
    def tol(self):
        """Desired convergence tolerance (relative)."""
        return self._tol

    @tol.setter
    def tol(self, value):
        if isinstance(value, float):
            if value > 0:
                self._tol = value
            else:
                raise ValueError("'tol' can't be zero or negative value.")
        else:
            raise ValueError("'tol' has to be a float.")

    @property
    def max_iter(self):
        """Maximum number of iterations."""
        return self._max_iter

    @max_iter.setter
    def max_iter(self, value):
        if isinstance(value, int):
            if value > 0:
                self._max_iter = value
            else:
                raise ValueError("'max_iter' can't be zero or negative value.")
        else:
            raise ValueError("'max_iter' has to be a integer.")

    @property
    def con_tol(self):
        """Desired threshold for the constraint violation."""
        return self._con_tol

    @con_tol.setter
    def con_tol(self, value):
        if isinstance(value, float):
            if value > 0:
                self._con_tol = value
            else:
                raise ValueError("'con_tol' can't be zero or negative value.")
        else:
            raise ValueError("'con_tol' has to be a float.")

    # -------------------------------------------------------------------------
    def __init__(self, name: str, server_url: str, tol: float = 1e-8,
                 max_iter: int = 3000, con_tol: float = 1e-4):
        super().__init__(name)

        self.server_url = server_url
        self.tol = tol
        self.max_iter = max_iter
        self.con_tol = con_tol

        self.test_connection()

    def test_connection(self):

        try:
            response = requests.get(self.server_url)
        except ConnectionError:
            raise ValueError("Couldn't connect to the server URL provided. "
                             "Make sure that the optimization server is "
                             "online and communicating properly.")
        else:
            if response.status_code != 200:
                raise ValueError("Connection to the server established. "
                                 "However, the server is unresponsive.")



from abc import ABC, abstractmethod
from typing import Union

import numpy as np
from scipy.linalg import norm
from scipy.spatial import cKDTree

from ..options import ProcedureOptions
from ..options.nlp import NLPOptions
from ..utils import _is_numeric_array_like
from .output import Report


class InfillProcedure(ABC):
    """Base class where the optimization through infill criteria is done.
    """

    @property
    def x(self) -> np.ndarray:
        """Design input samples."""
        return self._x

    @x.setter
    def x(self, value):
        if _is_numeric_array_like(value):
            self._x = np.asarray(value, dtype=float)
        else:
            raise ValueError("'x' has to be a numeric array.")

    @property
    def m(self) -> int:
        """Number of initial samples. (READONLY)"""
        return self._m

    @property
    def g(self) -> np.ndarray:
        """Observed constraint data."""
        return self._g

    @g.setter
    def g(self, value):
        if _is_numeric_array_like(value):
            # TODO: ensure it is not empty
            self._g = np.asarray(value, dtype=float)
        else:
            raise ValueError("'g' has to be a numeric array.")

    @property
    def f(self) -> np.ndarray:
        """Observed objective function data."""
        return self._f

    @f.setter
    def f(self, value):
        if _is_numeric_array_like(value):
            self._f = np.asarray(value, dtype=float).flatten()

        else:
            raise ValueError("'f' has to be a numeric array.")

    @property
    def model_function(self):
        """Rigorous model where the data will be sampled."""
        return self._model_function

    @model_function.setter
    def model_function(self, value):
        if callable(value) or value is None:
            self._model_function = value
        else:
            raise ValueError("'model_function' has to be a callable object.")

    @property
    def options(self):
        """Procedure options setup"""
        return self._options

    @options.setter
    def options(self, value: ProcedureOptions):
        if isinstance(value, ProcedureOptions):
            self._options = value
        else:
            raise ValueError("'options' has to be a valid options structure.")

    @property
    def nlp_options(self):
        """Non-Linear Programming solver options."""
        return self._nlp_options

    @nlp_options.setter
    def nlp_options(self, value: NLPOptions):
        if isinstance(value, NLPOptions):
            self._nlp_options = value
        else:
            raise ValueError("'nlp_options' has to be a valid NLP option "
                             "instance.")

    @property
    def lb(self):
        """Input design lower bound."""
        return self._lb

    @lb.setter
    def lb(self, value):
        if _is_numeric_array_like(value):
            self._lb = np.asarray(value, dtype=float).flatten()

        else:
            raise ValueError("'lb' has to be a numeric array.")

    @property
    def ub(self):
        """Input design upper bound."""
        return self._ub

    @ub.setter
    def ub(self, value):
        if _is_numeric_array_like(value):
            self._ub = np.asarray(value, dtype=float).flatten()

        else:
            raise ValueError("'ub' has to be a numeric array.")

    @property
    def report_options(self):
        """Optimization procedure options report (i.e. plot or return strings
        containing info about the procedure iterations)"""
        return self._report_options

    @report_options.setter
    def report_options(self, value):
        if isinstance(value, Report):
            self._report_options = value
        else:
            raise ValueError("'report_options' has to be a 'Report' object.")

    # -------------------------------------------------------------------------
    def __init__(self, x: np.ndarray, g: np.ndarray, f: np.ndarray,
                 model_function, lb: np.ndarray, ub: np.ndarray,
                 options: ProcedureOptions, nlp_options: NLPOptions,
                 report_options: Report):
        super().__init__()

        self.x = x
        self.g = g
        self.f = f
        self.model_function = model_function
        self.lb = lb
        self.ub = ub
        self.options = options
        self.nlp_options = nlp_options
        self.report_options = report_options

    @abstractmethod
    def check_setup(self):
        """ Checks if the problem data is ready to be optimized (i.e. check
        input, constraints and objective array dimensions, etc.)
        """
        # check input shapes
        n_x, d_x = self.x.shape
        n_g, d_g = self.g.shape
        n_f = self.f.size

        if self.f.ndim != 1:
            raise ValueError("The objective function sampled data has be a 1D "
                             "array.")

        if n_x != n_g or n_x != n_f:
            raise ValueError("'x', 'g' and 'f' must have the same number of "
                             "rows.")

        # if input dimensions are ok, store the number of initial points
        self._m = n_x

        d_lb = self.lb.size
        d_ub = self.ub.size

        if d_lb != d_ub:
            raise ValueError("'lb' and 'ub' must have the same size.")

        if d_x != d_lb or d_x != d_ub:
            raise ValueError("The number of input dimensions ('x') has to be "
                             "the same as 'lb' and 'ub'.")

        if np.any(self.lb >= self.ub):
            raise ValueError("'lb' elements must be greater than 'ub'.")

        # options checkup
        self.options.check_options_setup()

    @abstractmethod
    def optimize(self):
        """Performs the infill criteria optimization.
        This is an abstract method. Must be overriden.
        """

        self.check_setup()



from setuptools import setup, find_packages

with open("README.md", 'r') as fh:
    long_description = fh.read()

setup(name="surropt",
      version='0.0.10',
      description="Surrogate optimization toolbox for time consuming models",
      author="Felipe Souza Lima",
      author_email='feslima93@gmail.com',
      url='https://github.com/feslima/surropt',
      license='Apache License 2.0',
      packages=find_packages(
          'src', exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
      package_dir={'': 'src'},
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords="surrogate optimization, infill criteria optimization, blackbox optimization",
      install_requires=['numpy>=1.15', 'scipy>=1.2.0', 'pydace>=0.1.1',
                        'pydoe2>=1.2.1', 'requests>=2.20.1',
                        'colorama==0.4.1'],
      python_requires='>=3.5',
      classifiers=[
          "Programming Language :: Python :: 3.6",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Topic :: Scientific/Engineering"
      ])



import warnings
from copy import deepcopy

import numpy as np
from pydace import Dace
from scipy.linalg import norm

from ..core.nlp import optimize_nlp
from ..core.options.nlp import DockerNLPOptions, NLPOptions
from ..core.procedures import InfillProcedure
from ..core.procedures.output import Report
from ..core.utils import (get_samples_index, is_row_member,
                          point_domain_distance)
from .problem import CaballeroOptions, CaballeroReport, is_inside_hypercube


class Caballero(InfillProcedure):
    # TODO: (docstring) Add notes and example section. Also the optimization
    # problem description
    """Rigorous optimization of nonlinear programming problems (NLP) in which
    the objective function and/or some constraints are represented by noisy
    implicit black box functions. [1]_

    Parameters
    ----------
    x : np.ndarray
        Input variables of the Design of Experiments (DOE). Has to be a 2D
        array, with no duplicated rows.

    g : np.ndarray
        Constraints values (:math:`g(x) <= 0`) of the DOE. Has to be a 2D array
        with no duplicated rows.

    f : np.ndarray
        Objective function values of the DOE. Has to be a 1D array with no
        duplicated values.

    model_function : callable function
        Black box function callable object that evaluates the objective and
        constraints functions and returns them as dictionary object. See Notes
        on how to implement such function.

    lb : np.ndarray
        Lower bound of the input variables. Has to be 1D array with the number
        of elements being the same as the number of columns of `x`.

    ub : np.ndarray
        Upper bound of the input variables. Has to be 1D array with the number
        of elements being the same as the number of columns of `x`.

    regression : str
        Kriging mean regression model. Valid values are: 'poly0', 'poly1' and
        'poly2'.

    options : CaballeroOptions, optional
        Optimization procedure options. See `CaballeroOptions` for which
        parameters can be tweaked. Default is None, where a default instance of
        the options class is initialized. (The default values are described in
        the class `CaballeroOptions`).

    nlp_options : NLOPtions subclass, optional
        NLP solver options structure. Default is None, where a default instance
        of `DockerNLPOptions` is created with a server url pointing to the
        localhost address through port 5000 in order to the optimization
        problem to be solved by a flask application inside a WSL (Windows
        Subsystem for Linux) environment. This application uses IpOpt solver.

    report_options : Report (class or subclass), optional
        How to report the optimization procedure progress. Whether to print in
        terminal or plot each iteration. Default is set to print in terminal

    References
    ----------
    .. [1] J. A. Caballero, I. E. Grossmann. "An Algorithm for the Use of
        Surrogate Models in Modular Flowsheet Optimization". AIChE journal,
        vol. 54.10 (2008), pp. 2633-2650, 2008.
    """

    def __init__(self, x: np.ndarray, g: np.ndarray, f: np.ndarray,
                 model_function, lb: np.ndarray, ub: np.ndarray,
                 regression: str, options: CaballeroOptions = None,
                 nlp_options: NLPOptions = None,
                 report_options: Report = None):
        # kriging regression model
        self.regression = regression

        # proceed with default options for caballero procedure if none defined
        options = CaballeroOptions() if options is None else options

        # proceed with default options for NLP solver as docker server
        nlp_options = DockerNLPOptions(name='wsl-server',
                                       server_url='http://localhost:5000') \
            if nlp_options is None else nlp_options

        # proceed with default options for procedure report output
        report_options = CaballeroReport(terminal=True, plot=False) \
            if report_options is None else report_options

        # initialize mother class
        super().__init__(x=x, g=g, f=f, model_function=model_function, lb=lb,
                         ub=ub, options=options, nlp_options=nlp_options,
                         report_options=report_options)

        # perform a setup check
        self.check_setup()

        # initial surrogate construction
        self.build_surrogate(optimize=True)

    def check_setup(self):
        # perform basic checkup
        super().check_setup()

        # search for feasible points in the initial sampling
        best_feas_idx = self.search_best_feasible_index(g=self.g, f=self.f)

        if best_feas_idx is None:
            raise IndexError("No initial feasible point found. You need at "
                             "least one feasible case in the sample.")
        else:
            self.best_feas_idx = best_feas_idx

    def search_best_feasible_index(self, g: np.ndarray, f: np.ndarray) -> int:
        # get feasible points indexes for a given set of samples
        feas_idx = np.nonzero(np.all(g <= self.options.feasible_tol,
                                     axis=1))[0]

        if feas_idx.size == 0:
            # no feasible point found, return None
            return None
        else:
            # get feasible objective function values
            feas_obj = f[feas_idx]

            # get best feasible index
            # the np.argmin already handles multiple minima as first ocurrence
            return feas_idx[np.argmin(feas_obj)].item()

    def build_surrogate(self, optimize=False, x=None, f=None, g=None) -> None:
        """Builds the objective and constraints surrogates based on whether or
        not to optimize their hyperparameters.

        Parameters
        ----------
        optimize : bool, optional
            Whether or not optimize the hyperparameters, by default False.

        x : np.ndarray, optional
            The input data to be used. If None, use the initial values from
            construction.

        f : np.ndarray, optional
            The function objective data to be used. If None, use the initial
            values from construction.

        g : np.ndarray, optional
            The constraint data to be used. If None, use the initial values
            from construction.
        """
        x = self.x if x is None else x
        g = self.g if g is None else g
        f = self.f if f is None else f

        n = x.shape[1]
        q = g.shape[1]

        if optimize:
            # optimize hyperparameters

            # initial theta
            one = np.ones((n,))
            theta0 = one
            lob = 1e-12 * theta0
            upb = 1e2 * theta0

            surr_obj = Dace(regression=self.regression,
                            correlation='corrgauss')
            surr_obj.fit(S=x, Y=f, theta0=theta0, lob=lob, upb=upb)

            # constraints metamodel
            surr_con = []
            for i in range(q):
                obj_ph = Dace(regression=self.regression,
                              correlation='corrgauss')
                obj_ph.fit(S=x, Y=g[:, i], theta0=theta0, lob=lob, upb=upb)

                surr_con.append(obj_ph)

            # store models
            self.surr_obj = surr_obj
            self.surr_con = surr_con

        else:
            # do not optimize hyperparameters, use previous model data
            if not hasattr(self, 'surr_obj') or not hasattr(self, 'surr_con'):
                # no previous model found, use recursion to build it
                self.build_surrogate(optimize=True, x=x, g=g, f=f)

            else:
                # objective function
                self.surr_obj.fit(S=x, Y=f,
                                  theta0=self.surr_obj.theta.flatten())

                # constraints
                for idx, obj in enumerate(self.surr_con):
                    obj.fit(S=x, Y=g[:, idx], theta0=obj.theta.flatten())

    def optimize(self):
        super().optimize()

        # initialize counter and domains
        self.k, self.j = 1, 1
        lb, ub = self.lb, self.ub
        self.lbopt, self.hlb, self.dlb = deepcopy(lb), deepcopy(lb), \
            deepcopy(lb)
        self.ubopt, self.hub, self.dub = deepcopy(ub), deepcopy(ub), \
            deepcopy(ub)
        self.fun_evals, self.move_num, self.contract_num = 0, 0, 0

        # initial NLP solver estimate
        x0 = self.x[self.best_feas_idx, :].flatten()

        # create internal variables of caballero
        self._xlhs = deepcopy(self.x)
        self._gobs = deepcopy(self.g)
        self._fobs = deepcopy(self.f)

        # initialize xstar, gstar and fstar
        self._xstar = x0.reshape(1, -1)
        self._gstar = self._gobs[self.best_feas_idx, :].reshape(1, -1)
        self._fstar = self._fobs[self.best_feas_idx].flatten()

        # termination flag
        self.terminated = False

        # movement flag
        self._last_move = 'None'

        # print headers if terminal is specified
        rpt = self.report_options.print_iteration(movement=self._last_move,
                                                  iter_count=None,
                                                  x=x0.tolist(),
                                                  f_pred=None,
                                                  f_actual=None,
                                                  g_actual=None,
                                                  color_font=None,
                                                  header=True)

        while not self.terminated:
            sol = optimize_nlp(procedure=self, x=self._xlhs, g=self._gobs,
                               f=self._fobs, nlp_options=self.nlp_options,
                               x0=x0.tolist(), lb=self.lbopt.tolist(),
                               ub=self.ubopt.tolist())
            xjk, fjk, exitflag = sol['x'], sol['fval'], sol['exitflag']

            if self.fun_evals >= self.options.max_fun_evals:
                warnings.warn("Maximum number of function evaluations "
                              "achieved!")

                # search feasible indexes
                feas_idx = self.search_best_feasible_index(self._gobs,
                                                           self._fobs)
                if feas_idx is not None:
                    rpt = self.report_options
                    rpt.get_results_report(index=feas_idx,
                                           r=0.005, x=self._xlhs,
                                           f=self._fobs, lb=self.lb,
                                           ub=self.ub,
                                           fun_evals=self.fun_evals)

                # break loop
                self.terminated = True

                # store results as class variables
                self.xopt = self._xlhs[feas_idx, :].flatten()
                self.gopt = self._gobs[feas_idx, :].flatten()
                self.fopt = self._fobs[feas_idx]

            if not is_row_member(xjk, self._xlhs):
                sampled_results = self.sample_model(xjk)

                # update sample data
                self._xlhs = np.vstack((self._xlhs, xjk))
                self._gobs = np.vstack((self._gobs, sampled_results['g']))
                self._fobs = np.append(self._fobs, sampled_results['f'])

                self.fun_evals += 1

                # iteration display
                if exitflag < 1:
                    # infeasible
                    color_font = 'red'
                else:
                    # feasible
                    color_font = None

                max_feas = np.max(sampled_results['g'])
                self.report_options.print_iteration(movement=self._last_move,
                                                    iter_count=self.j,
                                                    x=xjk.tolist(),
                                                    f_pred=fjk,
                                                    f_actual=sampled_results['f'],
                                                    g_actual=max_feas,
                                                    color_font=color_font)
                self.report_options.plot_iteration()
            else:
                # couldn't improve from last iteration
                xjk = self.refine()
                x0 = deepcopy(xjk)

            # wheter or not to update kriging parameters after refinement phase
            optimize_hyp = True if self.j == 1 else False

            self.build_surrogate(x=self._xlhs, f=self._fobs, g=self._gobs,
                                 optimize=optimize_hyp)

            # Starting from xjk solve the NLP to get xj1k
            sol = optimize_nlp(procedure=self, x=self._xlhs, g=self._gobs,
                               f=self._fobs, nlp_options=self.nlp_options,
                               x0=xjk.tolist(), lb=self.lbopt.tolist(),
                               ub=self.ubopt.tolist())
            xj1k, fj1k, exitflag = sol['x'], sol['fval'], sol['exitflag']

            if point_domain_distance(xjk, xj1k, lb, ub) >= \
                    self.options.ref_tol:
                # there was improvement, keep going
                xjk = xj1k
                self.j += 1

            else:
                xjk = self.refine()
                x0 = deepcopy(xjk)

    def sample_model(self, x: np.ndarray):
        sampled_data = self.model_function(x)
        f = sampled_data['f']
        g = sampled_data['g']

        # TODO: implement other parameters capture in 'extras' key
        # TODO: output data sanitation (g, f, and extras)
        return {'g': g, 'f': f}

    def refine(self):
        # find best value inserted
        x_ins = self._xlhs[self.m:, :]
        g_ins = self._gobs[self.m:, :]
        f_ins = self._fobs[self.m:]

        best_idx = self.search_best_feasible_index(g_ins, f_ins)

        if best_idx is not None:
            # TODO: check for duplicated. If so, perform contraction on best
            self._xstar = np.vstack((self._xstar, x_ins[best_idx, :]))
            self._gstar = np.vstack((self._gstar, g_ins[best_idx, :]))
            self._fstar = np.append(self._fstar, f_ins[best_idx])

        else:
            # no best sampled value found in the inserted points, just insert
            # the best value in the initial sample (i.e. no improvement).
            # This block shouldn't be possible. The else is a "just in case".

            # TODO: perform a contraction on the best feasible point instead
            best_init = self.search_best_feasible_index(self.g, self.f)

            raise NotImplementedError("Perform a contraction move.")

        # select the best point to be centered
        best_star = self.search_best_feasible_index(self._gstar, self._fstar)

        xstark = self._xstar[best_star, :].flatten()

        if self.contract_num == 0:
            contract_factor = self.options.first_factor
        else:
            contract_factor = self.options.second_factor

        # refine the hypercube limits
        self.refine_hypercube(xstark, contract_factor)

        # search for best feasible point
        feas_idx = self.search_best_feasible_index(self._gobs,
                                                   self._fobs)

        # check for termination
        if point_domain_distance(self._xstar[-1, :],
                                 self._xstar[-2, :], self.lb, self.ub) <= \
                self.options.term_tol:

            xstark = self._xlhs[feas_idx, :].flatten()

            # check if at least a contraction was made
            if self.contract_num > 0:
                # if so, terminate the algorithm
                if feas_idx is not None:
                    rpt = self.report_options
                    rpt.get_results_report(index=feas_idx,
                                           r=0.005,
                                           x=self._xlhs,
                                           f=self._fobs,
                                           lb=self.lb,
                                           ub=self.ub,
                                           fun_evals=self.fun_evals)
                # break loop
                self.terminated = True

                # store results as class variables
                self.xopt = self._xlhs[feas_idx, :].flatten()
                self.gopt = self._gobs[feas_idx, :].flatten()
                self.fopt = self._fobs[feas_idx]

            else:
                # perform a large contraction if no contraction done
                self.refine_hypercube(xstark, contract_factor=0.9999)

        # update move counter and reset iteration counter
        self.k += 1
        self.j = 1

        return xstark

    def refine_hypercube(self, xstark: np.ndarray, contract_factor: float):
        d_range = self.hub - self.hlb

        # check if xstark is at domain bound
        if is_inside_hypercube(xstark, self.dlb, self.dub):
            # inside original domain
            if is_inside_hypercube(xstark, self.hlb, self.hub) and \
                norm(self.hub - self.hlb) / norm(self.dub - self.dlb) >= \
                    self.options.contraction_tol:
                # its inside hypercube, center and contract
                self._perform_contraction(xstark, contract_factor, d_range)

            else:
                # its at hypercube limit, center and move
                self._perform_move(xstark, d_range)

        else:
            # at domain limit
            if is_inside_hypercube(xstark, self.hlb, self.hub) and \
                norm(self.hub - self.hlb) / norm(self.dub - self.dlb) >= \
                    self.options.contraction_tol:
                # its inside hypercube, center and contract
                self._perform_contraction(xstark, contract_factor, d_range)

            else:
                # its at hypercube limit, center and move
                self._perform_move(xstark, d_range)

        # update optimization bounds and adjust them if needed
        self.lbopt = deepcopy(self.hlb)
        self.ubopt = deepcopy(self.hub)

        floor_lb = np.less(self.lbopt, self.dlb)
        if np.any(floor_lb):
            self.lbopt[floor_lb] = self.dlb[floor_lb]

        ceil_ub = np.greater(self.ubopt, self.dub)
        if np.any(ceil_ub):
            self.ubopt[ceil_ub] = self.dub[ceil_ub]

    def _perform_contraction(self, xstark, contract_factor, d_range):
        red_factor = (1 - contract_factor) * d_range / 2
        self.hlb = xstark - red_factor
        self.hub = xstark + red_factor

        self.contract_num += 1

        # inserted points
        x_ins = self._xlhs[self.m:, :]
        g_ins = self._gobs[self.m:, :]
        f_ins = self._fobs[self.m:]

        # each contraction discards the points outside new hypercube
        idx = get_samples_index(x_ins, self.hlb, self.hub)

        self._xlhs = np.vstack((self.x, x_ins[idx, :]))
        self._gobs = np.vstack((self.g, g_ins[idx, :]))
        self._fobs = np.append(self.f, f_ins[idx])

        self._last_move = 'Contraction'

    def _perform_move(self, xstark, d_range):
        red_factor = d_range / 2
        self.hlb = xstark - red_factor
        self.hub = xstark + red_factor

        self.move_num += 1

        self._last_move = 'Movement'



import os
from abc import ABC, abstractmethod
from string import Template

import numpy as np
from colorama import Fore
from scipy.linalg import norm
from scipy.spatial import cKDTree


class Report(ABC):
    # TODO: write class documentation
    def __init__(self, terminal=False, plot=False):
        super().__init__()

        self.terminal = terminal
        self.plot = plot

    @property
    def terminal(self):
        """Whether or not to print each iteration info as string."""
        return self._terminal

    @terminal.setter
    def terminal(self, value):
        if isinstance(value, bool):
            self._terminal = value
        else:
            raise ValueError("'terminal' property only accepts True or False.")

    @property
    def plot(self):
        """Whether or not to plot the iteration process."""
        return self._plot

    @plot.setter
    def plot(self, value):
        if isinstance(value, bool):
            self._plot = value
        else:
            raise ValueError("'plot' property only accepts True or False.")

    def build_iter_report(self, iter_count: int, x: list, f_pred: float,
                          f_actual: float, g_actual: float,
                          header=False, field_size: int = 10) -> str:
        # builds the string report for the current iteration. To integrate the
        # optimization procedure in other applicatins, you can extend
        # (override) this function.
        n_x = len(x)

        if header:
            mv_header = [" x" + str(i + 1) for i in range(n_x)]
            hd_temp = Template("{:$f_size}\t").substitute(f_size=field_size)
            str_arr = ['Iter'] + mv_header + \
                ['f_pred', 'f_actual', 'feasibility']
            arr_str = (hd_temp*len(str_arr)).format(*str_arr)

        else:
            i = str(iter_count)
            mv_arr = np.array(x)
            num_arr = np.append(x, np.array([f_pred, f_actual, g_actual]))
            f_temp = Template("{0: $f_size.4e}").substitute(f_size=field_size)
            formatter = {
                'float_kind': lambda x: f_temp.format(x)
            }
            # from https://bugs.python.org/issue24920: "for a GUI app the
            # Windows C runtime leaves the standard FILE streams uninitialized
            # to an invalid file descriptor (-1), so Python's sys.__stdout__
            # is None. That's why you get an AttributeError complaining that
            # NoneType (i.e. type(None)) has no attribute 'fileno'."

            # Attempt at fix for now:
            # try to get default terminal width, resort to 80 chars if it fails
            try:
                width = os.get_terminal_size()[0]
            except (OSError, ValueError):
                width = 80

            str_arr = np.array2string(num_arr, separator='\t',
                                      max_line_width=width,
                                      formatter=formatter)[1:-1]
            arr_temp = Template("{0:$f_size}\t{1}").substitute(
                f_size=field_size)
            arr_str = arr_temp.format(i, str_arr)

        return arr_str

    @abstractmethod
    def print_iteration(self, iter_count: int, x: list, f_pred: float,
                        f_actual: float, g_actual: float, header=False,
                        field_size: int = 10, color_font=None) -> None:
        # you have to override this function
        pass

    def plot_iteration(self):
        if self.plot:
            raise NotImplementedError("plot iteration not implemented!")

    def get_results_report(self, index: int, r: float, x: np.ndarray,
                           f: np.ndarray, lb: np.ndarray, ub: np.ndarray,
                           fun_evals: int) -> str:
        """Returns the results message report that contains info about the
        neighbourhood of `x` inside a specified domain (`lb` and `ub`).

        Parameters
        ----------
        index : int
            Row index of `x` that corresponds to the optimal point.
        r : float
            Radius percentage of the domain euclidian range (`lb` and `ub`) to
            search points near `x`.
        x : np.ndarray
            Sample input variables to be analysed (2D array).
        f : np.ndarray
            Sample observed objective function values (1D array, number of
            elements has to be the same as the number of rows in `x`).
        lb : np.ndarray
            Domain lower bound of the sample `x`. It is assumed that ALL the
            row of `x` are inside this domain.
        ub : np.ndarray
            Domain upper bound of the sample `x`. It is assumed that ALL the
            row of `x` are inside this domain.
        fun_evals : int
            Number of function evaluations needed to obtain the sample `x`

        Returns
        -------
        str
            Results report message to be printed in the terminal/cmd.
        """
        # search nearest points within r euclidian distance
        kdtree = cKDTree(data=x)
        euc_dom_rng = norm(ub - lb, ord=2)
        neigh_idx = kdtree.query_ball_point(x=x[index, :], r=r*euc_dom_rng)
        results_msg = ("\nBest feasible value found: {0:8.4f} at point\n"
                       "x = {1}\n"
                       "{2} points are within {3:.3%} euclidian range of this "
                       "point based on original domain.\n"
                       "Number of function evaluations: {4}")
        num_arr = np.array2string(x[index, :], precision=4, separator='\t',
                                  sign=' ')[1:-1]
        results_msg = results_msg.format(f[index],
                                         num_arr,
                                         len(neigh_idx),
                                         r, fun_evals)

        if self.terminal:
            # terminal asked, print directly
            print(Fore.RESET + results_msg)

        return results_msg
